// src/index.ts
var defaultObjectHash = (obj, context) => {
  return context.index.toString();
};
function generateJSONPatch(before, after, config = {}) {
  const { objectHash = defaultObjectHash, propertyFilter } = config;
  const patch = [];
  const hasPropertyFilter = typeof propertyFilter === "function";
  function compareArrays(leftArr, rightArr, path) {
    var _a;
    if (JSON.stringify(leftArr) === JSON.stringify(rightArr))
      return;
    const leftHashes = leftArr.map(
      (value, index) => objectHash(value, { side: "left", path, index })
    );
    const rightHashes = rightArr.map(
      (value, index) => objectHash(value, { side: "right", path, index })
    );
    let currentIndex = leftArr.length - 1;
    const targetHashes = [];
    for (let i = leftArr.length - 1; i >= 0; i--) {
      const newPathIndex = `${path}/${currentIndex--}`;
      const rightHashIndex = rightHashes.indexOf(leftHashes[i]);
      if (rightHashIndex >= 0) {
        compareObjects(newPathIndex, leftArr[i], rightArr[rightHashIndex]);
        targetHashes.unshift(leftHashes[i]);
      } else {
        patch.push({ op: "remove", path: newPathIndex });
      }
    }
    const toBeAddedHashes = rightHashes.filter(
      (hash) => !targetHashes.includes(hash)
    );
    currentIndex = targetHashes.length;
    for (const toBeAddedHash of toBeAddedHashes) {
      patch.push({
        op: "add",
        path: `${path}/${currentIndex++}`,
        value: rightArr[rightHashes.indexOf(toBeAddedHash)]
      });
      targetHashes.push(toBeAddedHash);
    }
    if ((_a = config.array) == null ? void 0 : _a.ignoreMove) {
      return;
    }
    for (let i = rightHashes.length - 1; i >= 0; i--) {
      const hash = rightHashes[i];
      const targetIndex = rightHashes.indexOf(hash);
      const currentIndex2 = targetHashes.indexOf(hash);
      if (currentIndex2 !== targetIndex) {
        patch.push({
          op: "move",
          from: `${path}/${currentIndex2}`,
          path: `${path}/${targetIndex}`
        });
        moveArrayElement(targetHashes, currentIndex2, targetIndex);
      }
    }
  }
  function compareObjects(path, leftJsonValue, rightJsonValue) {
    const isArrayAtTop = path === "" && [leftJsonValue, rightJsonValue].every(Array.isArray);
    if (isPrimitiveValue(leftJsonValue) || isPrimitiveValue(rightJsonValue)) {
      if (leftJsonValue !== rightJsonValue) {
        patch.push({ op: "replace", path, value: rightJsonValue });
      }
      return;
    }
    if (isArrayAtTop) {
      return compareArrays(leftJsonValue, rightJsonValue, "");
    }
    if ([leftJsonValue, rightJsonValue].some(Array.isArray)) {
      patch.push({ op: "replace", path, value: rightJsonValue });
      return;
    }
    for (const rightKey in rightJsonValue) {
      if (hasPropertyFilter && !propertyFilter(rightKey, { side: "right", path }))
        continue;
      let newPath = isArrayAtTop && path === "" ? `/${rightKey}` : `${path}/${rightKey}`;
      const leftValue = leftJsonValue[rightKey];
      const rightValue = rightJsonValue[rightKey];
      if (Array.isArray(leftValue) && Array.isArray(rightValue)) {
        compareArrays(leftValue, rightValue, newPath);
      } else if (isJsonObject(rightValue)) {
        if (isJsonObject(leftValue)) {
          compareObjects(newPath, leftValue, rightValue);
        } else if (leftJsonValue.hasOwnProperty(rightKey)) {
          patch.push({ op: "replace", path: newPath, value: rightValue });
        } else {
          patch.push({ op: "add", path: newPath, value: rightValue });
        }
      } else if (!leftJsonValue.hasOwnProperty(rightKey)) {
        patch.push({ op: "add", path: newPath, value: rightValue });
      } else if (leftValue !== rightValue) {
        patch.push({ op: "replace", path: newPath, value: rightValue });
      }
    }
    for (const leftKey in leftJsonValue) {
      if (!leftJsonValue.hasOwnProperty(leftKey) || hasPropertyFilter && !propertyFilter(leftKey, { side: "left", path }))
        continue;
      if (!rightJsonValue.hasOwnProperty(leftKey)) {
        let newPath = isArrayAtTop && path === "" ? `/${leftKey}` : `${path}/${leftKey}`;
        patch.push({ op: "remove", path: newPath });
      }
    }
  }
  compareObjects("", before, after);
  return [...patch];
}
function isPrimitiveValue(value) {
  return value === void 0 || value === null || typeof value === "undefined" || typeof value === "string" || typeof value === "number" || typeof value === "boolean";
}
function isJsonObject(value) {
  return (value == null ? void 0 : value.constructor) === Object;
}
function moveArrayElement(array, from, to) {
  array.splice(to, 0, array.splice(from, 1)[0]);
}
function pathInfo(path) {
  const segments = path.split("/");
  const length = segments.length;
  const last = segments[length - 1];
  return { segments, length, last };
}
export {
  defaultObjectHash,
  generateJSONPatch,
  pathInfo
};
